unit Creature;

interface

uses
  Windows, SysUtils, Classes, Graphics, Math, Types;

(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        TCreature
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)
type
    TCreature = class
    protected
        fColor : TColor;    //  ̐F
        fMagnification: double; //  ̂̑傫
    public
        fX  : double;       //  EɂXW
        fY  : double;       //  EɂYW
        fZ  : double;       //  EɂZW
        fAngle : double;    //  Ă
        fSpeed : double;    //  1̈ړňړ鋗
        fLife  : integer;   //  

        fVisibleAngle   : double;
        fVisibleDistance: double;

        fEscape: integer;
        fEscapeAngle: double;

        fDrawAngle: Boolean;
        fDrawConnection: Boolean;

        fTargetCreature: TCreature;     //ǂΏۂ̃IuWFNg

        constructor Create();
        destructor  Destroy(); override;
        procedure   mMove();   virtual;             //  ړ
        procedure   mDraw();   virtual;             //  `
        procedure   mDrawConnection();
        procedure   mDrawAngle();
        procedure   mFindTarget();  virtual;
        procedure   mNextState();   virtual;
    end;

(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        TCreatureList
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)
type
    TCreatureList = class(TList)
    protected
        function  Get(Index: Integer): TCreature;
        procedure Put(Index: Integer; const Value: TCreature);
    public
        property Items[Index: Integer]: TCreature read Get write Put; default;
        destructor Destroy; override;
    end;
(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        TWorld
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)
type
    TWorld = class
    public
        fWidth : integer;
        fHeight: integer;
        fCreatureList  : TCreatureList;
        fCanvas : TCanvas;

        fDrawBackGround: Boolean;

        constructor Create(pWidth, pHeight: integer; pCanvas: TCanvas; pBkFlg: Boolean);
        destructor  Destroy(); override;
        procedure   mCloudCenter(var pX, pY: double);   // Wc̒S
        procedure   mUpdate();
    end;

const
    PI  = System.Pi;
    PI2 = 2 * PI;
    MAX_LIFE      = 100;
var
    aWorld: TWorld;

implementation

uses
    Common;

(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        TCreatureList
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)
//  ==========================================================
function TCreatureList.Get(Index: Integer): TCreature;
begin
    Result :=  TCreature( inherited Get(Index) );
end;
//  ==========================================================
procedure TCreatureList.Put(Index: Integer; const Value: TCreature);
begin
    inherited Put( Index, Value );
end;
//  ==========================================================
destructor TCreatureList.Destroy;
var
    i: Integer;
begin
    for i := 0 to Self.Count - 1 do Self.Items[i].Free;
    inherited Destroy;
end;

(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        TWorld
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)
//  ==========================================================
constructor TWorld.Create(pWidth, pHeight: integer; pCanvas: TCanvas; pBkFlg: Boolean);
begin
    inherited Create();
    fCreatureList := TCreatureList.Create;
    fWidth  := pWidth;
    fHeight := pHeight;
    fCanvas := pCanvas;
    fDrawBackGround := pBkFlg;

    with fCanvas do begin
        Pen.Color := clBlack;
        with Font do begin
            Color := clGray;
            Size  := 150;
            Charset := ANSI_CHARSET;
            Name    := 'Times New Roman';
            Style   := Style + [fsbold];
        end;
    end;
end;
//  ==========================================================
destructor  TWorld.Destroy;
begin
    fCreatureList.Free;
    inherited;
end;
//  ==========================================================
procedure TWorld.mCloudCenter(var pX, pY: double);   // Wc̒S
var
    i: integer;
    x, y: double;
begin
    x := 0;
    y := 0;
    for i := 0 to fCreatureList.Count - 1 do begin
        x := x + fCreatureList.Items[i].fX;
        y := y + fCreatureList.Items[i].fY;
    end;
    pX := x / fCreatureList.Count;
    pX := y / fCreatureList.Count;
end;
//  ==========================================================
procedure TWorld.mUpdate();   // `XV
var
    i: integer;
    Num: string;
    aCreature: TCreature;
begin

    fCanvas.Brush.Color := clWhite;
    fCanvas.FillRect(Bounds(0, 0, fWidth, fHeight));

    if fDrawBackGround then begin
        Num := IntToStr(fCreatureList.Count);
        fCanvas.TextOut(
            (fWidth  - fCanvas.TextWidth(Num)) div 2,
            (fHeight - fCanvas.TextHeight(Num)) div 2,
            Num
        );

        i := 0;
        fCanvas.Pen.Color := clGray;
        while i < fHeight do begin
            fCanvas.Rectangle(0, i, fWidth, i + 1);
            Inc(i, 5);
        end;
        i := 1;
        fCanvas.Pen.Color := clWhite;
        while i < fHeight do begin
            fCanvas.Rectangle(0, i, fWidth, i + 1);
            Inc(i, 5);
        end;
    end;

    for i := fCreatureList.Count - 1 downto 0 do begin
        aCreature := fCreatureList.Items[i];
        aCreature.mNextState;
        if aCreature.fLife > MAX_LIFE then aCreature.Free;
    end;

    for i := 0 to fCreatureList.Count - 1 do begin
        fCreatureList.Items[i].mMove;
        fCreatureList.Items[i].mDrawConnection;
        fCreatureList.Items[i].mDrawAngle;
    end;

    for i := 0 to fCreatureList.Count - 1 do begin
        fCreatureList.Items[i].mDraw;
    end;

end;
(*
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        TCreature
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*)
//  ==========================================================
constructor TCreature.Create();
const
    MAX_SPEED = 5;
begin
    inherited;
    aWorld.fCreatureList.Add(Self);
    fColor           := clMenu;
    fAngle           := PI2 * Random;
    fDrawAngle       := True;
    fDrawConnection  := True;
    fMagnification   := 0.25;
    fVisibleDistance := Random(50) + 200;
    fVisibleAngle    := PI  * Urand(0.6, 0.9);
    fSpeed           := MAX_SPEED;
    fEscape          := 0;
    fLife            := 0;
end;
//  ==========================================================
destructor  TCreature.Destroy;
begin
    aWorld.fCreatureList.Remove(Self);
    inherited;
end;
//  ==========================================================
procedure TCreature.mMove();
var
    Cos: Extended;
    Sin: Extended;
begin
    //  ʒuړ
    SinCos(fAngle, Sin, Cos);
    fX := fX + Cos * fSpeed;
    fY := fY + Sin * fSpeed;
    //@Ê͂ݏo(XW)
    if (fX < 0) then begin
        fX := fX + aWorld.fWidth;
    end else if (fX > aWorld.fWidth) then begin
        fX := fX - aWorld.fWidth;
    end;
    //@Ê͂ݏo(YW)
    if (fY < 0) then begin
        fY := fY + aWorld.fHeight;
    end else if (fY > aWorld.fHeight) then begin
        fY := fY - aWorld.fHeight;
    end;

end;

procedure TCreature.mDraw();
begin
    with aWorld.fCanvas do begin
        Pen.Color   := fColor;
        Brush.Color := fColor;
        Ellipse(Round(fX - 5), Round(fY - 5), Round(fX + 5), Round(fY + 5));
    end;
end;

procedure TCreature.mDrawAngle();
var
    Sin, Cos: Extended;
    dXA, dYA, dThetaA : double;
    dXC, dYC, dThetaC : double;
begin

    if not fDrawAngle then Exit;

    //S猩ĎE[
    dThetaA := fAngle - fVisibleAngle * 0.5;
    SinCos(dThetaA, Sin, Cos);
    dXA := fVisibleDistance * Cos;
    dYA := fVisibleDistance * Sin;
    //S猩Ď썶[
    dThetaC := fAngle + fVisibleAngle * 0.5;
    SinCos(dThetaC, Sin, Cos);
    dXC := fVisibleDistance * Cos;
    dYC := fVisibleDistance * Sin;

    //撣ĕ`
    with aWorld.fCanvas do begin

        Pen.Color   := fColor;
        Brush.Color := fColor;
        Brush.Style := bsClear;

        Pie(Round(fX - fVisibleDistance) , Round(fY - fVisibleDistance),
            Round(fX + fVisibleDistance) , Round(fY + fVisibleDistance),
            Round(fX + dXC), Round(fY + dYC),
            Round(fX + dXA), Round(fY + dYA)
            );

    end;
end;

//  -------------------------------------------------------------
//  ^[Qbg͈͓̔(猩)ɂ邩ǂ`FbN
function TargetInDistance(Hunter, Target: TCreature; MaxDistance: double; var Distance2: double): Boolean;
var
    DeltaX, DeltaY: double;
begin
    DeltaX      := Hunter.fX - Target.fX;
    DeltaY      := Hunter.fY - Target.fY;
    Distance2   := DeltaX * DeltaX + DeltaY * DeltaY;
    Result      := (MaxDistance > Distance2);
end;

function GetAngle(vx, vy: double): double;
var
    a: double;
begin
    //  _ƂāAxɂꍇ
    if (vy = 0) then begin
        if (vx < 0) then Result := PI else Result := 0;
    //  xł͂Ȃꍇ
    end else begin
        a := ArcCos(vx / sqrt(vx * vx + vy * vy));
        if (vy > 0) then Result := a else Result := PI2 - a;
    end;
end;

function AngleDifference(t, s: double): double;
var
    a: double;
begin
    a := t - s;
    if a > PI then begin
        Result := a - PI2;
    end else if a < -PI then begin
        Result := a + PI2;
    end else begin
        Result := a;
    end;
end;

//  -------------------------------------------------------------
//  ^[Qbg͎EɓĂȂƃ_
function TargetInSight(Hunter, Target: TCreature): Boolean;
var
    DeltaX, DeltaY: double;
    b: double;
begin
    DeltaX      := Hunter.fX - Target.fX;
    DeltaY      := Hunter.fY - Target.fY;
    b := AngleDifference(GetAngle(DeltaX, DeltaY), Hunter.fAngle);
//    Result := ( (-0.5 * Hunter.fVisibleAngle < b) and (b < 0.5 * Hunter.fVisibleAngle) );
    Result := not ( (b <= -0.5 * Hunter.fVisibleAngle) or (0.5 * Hunter.fVisibleAngle <= b) );
end;

//  ^[Qbg͕ĂKv
function AngleSimilarity(Hunter, Target: TCreature): Boolean;
var
    b: double;
begin
    b := AngleDifference(Hunter.fAngle, Target.fAngle);
    Result := ( (b > -0.6 * PI) or (b < 0.6 * PI) );
end;

procedure TCreature.mFindTarget();
var
    i: integer;
    Target    : TCreature;
    FindTarget: TCreature;
    MaxDistance: double;
    d: double;
begin

    FindTarget := nil;
    //  ő勖e
    MaxDistance := fVisibleDistance * fVisibleDistance;

    for i := 0 to aWorld.fCreatureList.Count - 1 do begin
        Target := aWorld.fCreatureList.Items[i];
        //͂񏜊O
        if Target = Self then Continue;
        //قȂ^CvO
        if Self.ClassType <> Target.ClassType then Continue;
        //IɌꏊǂ
        if not TargetInDistance(Self, Target, MaxDistance, d) then Continue;
        //͈̔͂ɓĂ邩ǂ
        if not TargetInSight(Self, Target) then Continue;
        //Ă邩ǂ
        if not AngleSimilarity(Self, Target) then Continue;

        FindTarget  := Target;
        MaxDistance := d;
    end;
    Self.fTargetCreature := FindTarget;
end;

procedure TCreature.mNextState();
const
    FLOCK_WEIGHT  = 0.06;       //  H
    CENTER_WEIGHT = 0.007;      //  H
    TARGET_WEIGHT = 0.1;        //  H
    MIMIC_WEIGHT  = 0.1;        //  H
    ACCELERATION  = 1.3;        //  x@H
    MAX_SPEED     = 14.0;
var
    x, y, a, s, vx, vy: double;
    CloudCenterX, CloudCenterY: double;
begin

    Inc(fLife);
    if fLife > MAX_LIFE * 3 div 4 then begin
        fColor := clRed;
    end;

    x := fX;
    y := fY;
    a := fAngle;
    s := fSpeed;

    if fEscape = 0 then begin

        mFindTarget();

        //  ǂ鑊肪Ȃ
        if fTargetCreature = nil then begin
            aWorld.mCloudCenter(CloudCenterX, CloudCenterY);
            a := a + FLOCK_WEIGHT * AngleDifference(GetAngle(CloudCenterX - x, CloudCenterY - y), a);
            s := (1 - 0.7 * Random) * MAX_SPEED;

        //  ǂ鑊肪鎞
        end else begin
            vx := fTargetCreature.fX - x;
            vy := fTargetCreature.fY - y;

            a := a + CENTER_WEIGHT * AngleDifference(GetAngle(aWorld.fWidth * 0.5 - x, aWorld.fHeight * 0.5 - y), a);
            a := a + TARGET_WEIGHT * AngleDifference(GetAngle(vx, vy), a);
            a := a + MIMIC_WEIGHT  * AngleDifference(fTargetCreature.fAngle, a);

            if (vx * vx + vy + vy < fVisibleDistance * fVisibleDistance) then s := 3 + 7 * Random;

            if (s > MAX_SPEED) then s := MAX_SPEED
            else if (s < 1)    then s := 1;

            //Xяo
            if Random < 0.002 then begin
                s := MAX_SPEED;
                if (aWorld.fCreatureList.IndexOf(fTargetCreature) = 0) then fEscapeAngle := 0
                else if (AngleDifference(GetAngle(vx, vy), a) < 0) then fEscapeAngle := 0.1 * PI
                else                                                    fEscapeAngle := -0.07 * PI;

                fTargetCreature := nil;
                fEscape := 9;
            end;

        end;

    end else begin
        a := a + fEscapeAngle;
        Dec(fEscape);
    end;

    if (a < 0) then
        a := a + PI2
    else if (a >= PI2) then
        a := a - PI2;

    fAngle  := a;
    fSpeed  := s;

end;

procedure TCreature.mDrawConnection();
begin
    if not fDrawConnection then Exit; 
    if fTargetCreature = nil then Exit;
    aWorld.fCanvas.Pen.Color := clRed;
    aWorld.fCanvas.MoveTo(Round(fX), Round(fY));
    aWorld.fCanvas.LineTo(Round(fTargetCreature.fX), Round(fTargetCreature.fY));
end;

end.
